@@ -23,6 +23,7 @@
 #include <string.h>
 #include <fcntl.h>
 #include <getopt.h>
+#include <fnmatch.h>
 #include <glib.h>
 
 #include "smd-config.h"
@@ -181,6 +182,10 @@ STATIC time_t lastcheck;
 // program options
 STATIC int verbose;
 STATIC int dry_run;
+STATIC int only_list_subfolders;
+STATIC int only_generate_symlinks;
+STATIC int n_excludes;
+STATIC char **excludes;
 
 // ============================ helpers =====================================
 
@@ -208,14 +213,15 @@ STATIC void assert_all_are(
 	int c, rc;
 	VERBOSE(input, "Asserting all input paths are: %s\n", description);
 	for(c = 0; c < argc; c++) { 
-		rc = stat(argv[c], &sb);
+		const char * argv_c = txtURL(argv[c], tmpbuff_5);
+		rc = stat(argv_c, &sb);
 		if (rc != 0) {
-			ERROR(stat,"unable to stat %s\n",argv[c]);
+			ERROR(stat,"unable to stat %s\n",argv_c);
 		} else if ( ! predicate(sb) ) {
 			ERROR(stat,"%s in not a %s, arguments must be omogeneous\n",
-				argv[c],description);
+				argv_c,description);
 		}
-		VERBOSE(input, "%s is a %s\n", argv[c], description);
+		VERBOSE(input, "%s is a %s\n", argv_c, description);
 	}
 }
 #define ASSERT_ALL_ARE(what,v,c) assert_all_are(what,tostring(what),v,c)
@@ -629,27 +635,55 @@ err_alloc_cleanup:
 	
 // recursively analyze a directory and its sub-directories
 STATIC void analyze_dir(const char* path){
-	DIR* dir = opendir(path);
+	DIR* dir;
 	struct dirent *dir_entry;
+	int inside_cur_or_new = 0;
+	int i, rc;
 
+	// skip excluded paths
+	for(i = 0; i < n_excludes; i++){
+		if ( (rc = fnmatch(excludes[i], path, 0)) == 0 ) {
+			VERBOSE(analyze_dir,
+				"skipping '%s' because excluded by pattern '%s'\n",
+				path, excludes[i]);
+			return;
+		}
+		if ( rc != FNM_NOMATCH ){
+			ERROR(fnmatch,"processing pattern '%s': %s",excludes[i],
+				strerror(errno))
+		}
+	}
+
+	// detect if inside cur/ or new/
+#ifdef __GLIBC__
+	const char* bname = basename(path);
+#else
+	gchar* bname = g_path_get_basename(path);
+#endif
+	if ( !strcmp(bname,"cur") || !strcmp(bname,"new") ) {
+		inside_cur_or_new = 1;
+		if ( only_list_subfolders ) {
+				fprintf(stdout, "%s\n", path);
+				return;
+		}
+	}
+#ifndef __GLIBC__
+	g_free(bname);
+#endif
+
+	dir = opendir(path);
 	if (dir == NULL) ERROR(opendir, "Unable to open directory '%s'\n", path);
 
 	while ( (dir_entry = readdir(dir)) != NULL) {
-		if (DT_REG == dir_entry->d_type){
-#ifdef __GLIBC__ 
-			const char* bname = basename(path);	
-#else
-			gchar* bname = g_path_get_basename(path);	
-#endif
-			if ( !strcmp(bname,"cur") || !strcmp(bname,"new")) {
+		if (DT_REG == dir_entry->d_type) {
+			if ( inside_cur_or_new && !only_list_subfolders ) {
 				analyze_file(path,dir_entry->d_name);
-			} else
+			} else {
 				VERBOSE(analyze_dir,"skipping '%s/%s', outside maildir\n",
 					path,dir_entry->d_name);
-#ifndef __GLIBC__ 
-			g_free(bname);
-#endif
-		} else if (DT_DIR == dir_entry->d_type && 
+			}
+		} else if ((DT_DIR == dir_entry->d_type ||
+					DT_LNK == dir_entry->d_type) &&
 				strcmp(dir_entry->d_name,"tmp") &&
 				strcmp(dir_entry->d_name,".") &&
 				strcmp(dir_entry->d_name,"..")){
@@ -660,15 +694,17 @@ STATIC void analyze_dir(const char* path){
 			free(newdir);
 		}
 	}
+	closedir(dir);
 }
 
 STATIC void analyze_dirs(char* paths[], int no){
 	int i;
 	for(i=0; i<no; i++){
 		// we remove a trailing '/' if any 
-		char *data = paths[i];
+		char *data = strdup(txtURL(paths[i],tmpbuff_5));
 		if (data[strlen(data)-1] == '/') data[strlen(data)-1] = '\0';
 		analyze_dir(data);
+		free(data);
 	}
 }
 
@@ -734,34 +770,41 @@ STATIC void extra_sha_file(const char* file, int suddenly_flush) {
 
 #define OPT_MAX_MAILNO 300
 #define OPT_DB_FILE    301
+#define OPT_EXCLUDE    302
 
 // command line options
 STATIC struct option long_options[] = {
-	{"max-mailno", 1, NULL, OPT_MAX_MAILNO},
-	{"db-file"   , 1, NULL, OPT_DB_FILE}, 
-	{"verbose"   , 0, NULL, 'v'},
-	{"dry-run"   , 0, NULL, 'd'},
-	{"help"      , 0, NULL, 'h'},
-	{NULL        , 0, NULL, 0}, 
+	{"max-mailno", required_argument, NULL, OPT_MAX_MAILNO},
+	{"db-file"   , required_argument, NULL, OPT_DB_FILE},
+	{"exclude"   , required_argument, NULL, OPT_EXCLUDE},
+	{"list"      , no_argument      , NULL, 'l'},
+	{"symlink"   , no_argument      , NULL, 's'},
+	{"verbose"   , no_argument      , NULL, 'v'},
+	{"dry-run"   , no_argument      , NULL, 'd'},
+	{"help"      , no_argument      , NULL, 'h'},
+	{NULL        , no_argument      , NULL, 0},
 };
 
 // command line options documentation
 STATIC const char* long_options_doc[] = {
-	"Estimation of max mail message number (defaults to the" 
-		"\n                        " 
-		"number of messages in the db-file + 1000 or "
-		tostring(DEFAULT_MAIL_NUMBER) 
-		"\n                        " 
-		"if there is no db-file). You may want to decrease it"
-		"\n                        " 
-		"for the first run on small systems. It is anyway"
-		"\n                        " 
-		"increased automatically when needed", 
-	"Name of the cache for the endpoint (default db.txt)",
-	"Increase program verbosity (printed on stderr, short -v)", 
-	"Do not generate a new db file (short -d)",
-	"This help screen", 
-	NULL
+	" number Estimation of max mail message number (defaults to the"
+				"\n                      "
+				"number of messages in the db-file + 1000 or "
+				tostring(DEFAULT_MAIL_NUMBER)
+				"\n                      "
+				"if there is no db-file). You may want to decrease it"
+				"\n                      "
+				"for the first run on small systems. It is anyway"
+				"\n                      "
+				"increased automatically when needed",
+	"path      Name of the cache for the endpoint (default db.txt)",
+	"glob      Exclude paths matching the given glob expression",
+			"Only list subfolders (short -l)",
+			"Symbolic Link generation mode (short -s)",
+			"Increase program verbosity (printed on stderr, short -v)",
+			"Do not generate a new db file (short -d)",
+			"This help screen",
+			NULL
 };
 
 // print help and bail out
@@ -771,8 +814,12 @@ STATIC void help(char* argv0){
 
 	fprintf(stdout,"\nUsage: %s [options] (dirs...|fifo)\n",bname);
 	for (i=0;long_options[i].name != NULL;i++) {
-		fprintf(stdout,"  --%-20s%s\n",
-			long_options[i].name,long_options_doc[i]);
+		if ( long_options[i].has_arg == required_argument )
+			fprintf(stdout,"  --%-8s%s\n",
+				long_options[i].name,long_options_doc[i]);
+		else
+			fprintf(stdout,"  --%-18s%s\n",
+				long_options[i].name,long_options_doc[i]);
 	}
 	fprintf(stdout,"\n\
 If paths is a single fifo, %s reads from it file names and outputs the\n\
@@ -809,7 +856,7 @@ int main(int argc, char *argv[]) {
 	g_assert(MAIL(GPTR(1)) == 1);
 
 	for(;;) {
-		c = getopt_long(argc, argv, "vhd", long_options, &option_index);
+		c = getopt_long(argc, argv, "vhdls", long_options, &option_index);
 		if (c == -1) break; // no more args
 		switch (c) {
 			case OPT_MAX_MAILNO:
@@ -818,12 +865,23 @@ int main(int argc, char *argv[]) {
 			case OPT_DB_FILE:
 				dbfile = strdup(optarg);
 			break;
+			case OPT_EXCLUDE:
+				excludes = realloc(excludes, sizeof(char*) * (n_excludes + 1));
+				excludes[n_excludes] = strdup(txtURL(optarg,tmpbuff_5));
+				n_excludes++;
+			break;
 			case 'v':
 				verbose = 1;
 			break;
 			case 'd':
 				dry_run = 1;
 			break;
+			case 'l':
+				only_list_subfolders = 1;
+			break;
+			case 's':
+				only_generate_symlinks = 1;
+			break;
 			case 'h':
 				help(argv[0]);
 				exit(EXIT_SUCCESS);
@@ -841,7 +899,7 @@ int main(int argc, char *argv[]) {
 	}
 
 	// remaining args is the dirs containing the data or the files to hash
-	data = argv[optind];
+	data = strdup(txtURL(argv[optind],tmpbuff_5));
 
 	// check if data is a directory or a regular file
 	c = stat(data, &sb);
@@ -853,30 +911,62 @@ int main(int argc, char *argv[]) {
 			ERROR(fopen,"unable to open fifo %s\n",data);
 			exit(EXIT_FAILURE);
 		}
-		while (!feof(in)) {
-			char name[MAX_EMAIL_NAME_LEN];	
-			if(fgets(name,MAX_EMAIL_NAME_LEN,in) != NULL){
-				size_t len = strlen(name);
-				if (len > 0 && name[len-1] == '\n') name[len-1]='\0';
-				extra_sha_file(txtURL(name,tmpbuff_5),1);
+		if ( only_generate_symlinks ) {
+			/* symlink */
+			char src_name[MAX_EMAIL_NAME_LEN];
+			char tgt_name[MAX_EMAIL_NAME_LEN];
+			while (!feof(in)) {
+				if(fgets(src_name,MAX_EMAIL_NAME_LEN,in) != NULL &&
+				   fgets(tgt_name,MAX_EMAIL_NAME_LEN,in) != NULL) {
+					size_t src_len = strlen(src_name);
+					size_t tgt_len = strlen(tgt_name);
+					if (src_len > 0 && src_name[src_len-1] == '\n')
+						src_name[src_len-1]='\0';
+					if (tgt_len > 0 && tgt_name[tgt_len-1] == '\n')
+						tgt_name[tgt_len-1]='\0';
+					gchar* dir_tgt = g_path_get_dirname(tgt_name);
+					g_mkdir_with_parents(dir_tgt, 0770);
+					if ( symlink(src_name, tgt_name) != 0 ){
+						ERROR(symlink,"unable to symlink %s to %s: %s\n",
+							src_name, tgt_name, strerror(errno));
+						exit(EXIT_FAILURE);
+					}
+					g_free(dir_tgt);
+				}
+			}
+		} else {
+			/* sha1 */
+			while (!feof(in)) {
+				char name[MAX_EMAIL_NAME_LEN];
+				if(fgets(name,MAX_EMAIL_NAME_LEN,in) != NULL){
+					size_t len = strlen(name);
+					if (len > 0 && name[len-1] == '\n') name[len-1]='\0';
+					extra_sha_file(name,1);
+				}
 			}
 		}
 		exit(EXIT_SUCCESS);
 	} else if ( ! S_ISDIR(sb.st_mode) ) {
 		ERROR(stat, "given path is not a fifo nor a directory: %s\n",data);
 	}
+	free(data);
 	
 	// regular case, hash the content of maildirs rooted in the 
 	// list of directories specified at command line
 	ASSERT_ALL_ARE(directory, &argv[optind], argc - optind);
 
+	if ( only_list_subfolders ) {
+		analyze_dirs(&argv[optind], argc - optind);
+		exit(EXIT_SUCCESS);
+	}
+
 	// allocate memory
 	setup_globals(dbfile, mailno, filenamelen);
 
 	load_db(dbfile);
 
 	bigbang = time(NULL);
-	analyze_dirs(&argv[optind],argc - optind);
+	analyze_dirs(&argv[optind], argc - optind);
 
 	generate_deletions();
 
